import numpy as np

# Universal filter class
class Filter:
    N = 0
    mu = 0
    h = np.array(float)
    x_vec = np.array(float)    

    def __init__(self, N, mu):

        self.N = N
        self.mu = mu
        self.h = np.zeros((N,))
        self.x_vec = np.zeros((N,))        

    def len(self):
        
        return self.N

    def filter(self, x):

        out = np.convolve(self.h,x)
        return out[0:len(x)]

    # Manages input vector for update
    def x_vec_update(self,x):

        # x is scalar
        self.x_vec[1:] = self.x_vec[0:-1]
        self.x_vec[0] = x

    # Adaptation method
    # May be called with vectors
    def adapt(self,x,d):

        # Sample-wise adaptation
        if np.isscalar(x):
            self.h = self.h + self.mu * self.delta(x,d)
        else:

            for i in range(len(x)):
                # Universal iterative update equation
                self.h = self.h + self.mu * self.delta(x[i],d[i])
    
    # To be implemented in derived classes
    def delta(self,x,d):
        return 0

# Least mean sqaure algoritm
class LMS(Filter):

    # LMS correction term
    def delta(self, x, d):

        # Prepare quantities
        self.x_vec_update(x)
        e = d - np.inner(self.x_vec,self.h)

        return e * self.x_vec

# Normalized least mean square algorithm
class NLMS(Filter):

    # NLMS correction term
    def delta(self,x,d):

        # Prepare quantities
        self.x_vec_update(x)
        e = d - np.inner(self.x_vec,self.h)

        return e / np.inner(self.x_vec,self.x_vec) * self.x_vec 

# Affine projection algorithm
class Affine(Filter):
    x_mat = np.array(float)
    d_vec = np.array(float) 

    def __init__(self,N,M,mu):

        Filter.__init__(self,N,mu)

        # Additinal quantaties
        self.M = M
        self.x_mat = np.zeros((N,M))
        self.d_vec = np.zeros((M,))
        self.eps = np.identity(M)*0.000000000001
    
    # Manages input matrix for update
    def x_mat_update(self,x): 

        # Utilize vector update function
        self.x_vec_update(x)

        self.x_mat[:,1:] = self.x_mat[:,:-1]
        self.x_mat[:,0] = self.x_vec

    # Affine projection correction term
    def delta(self,x,d):

        # Input update
        self.x_mat_update(x)

        # Output update
        self.d_vec[1:] = self.d_vec[0:-1]
        self.d_vec[0] = d

        # Prepare other quantities
        e = self.d_vec - np.matmul(np.transpose(self.x_mat),self.h)
        inverse = np.linalg.inv(np.transpose(self.x_mat)@self.x_mat + self.eps)

        return self.x_mat @ inverse @ e
