import numpy as np
import soundfile as sf
import matplotlib.pyplot as plt

# Our own filter class
import adaptivefilters_solution as af

# Read in audiofiles
x, samplerate = sf.read('wn_10s.wav')
n, samplerate = sf.read('meas_noise.wav')
x_power = np.mean(x**2)
n_power = np.mean(n**2)
l = len(x)

# Read in true impulse response (length 256)
h, samplerate = sf.read('h_256.wav')
h_norm = np.linalg.norm(h)**2
N = 256

# Simulate system output 
d = np.convolve(x,h)
d = d[0:l]

# Scale measurement noise
# SNR with respect to excitation
SNR = 30
SNR_lin = 10**(SNR/10)
n_scaled = n*np.sqrt(x_power/n_power/SNR_lin)

# Disturbed observed signal
y = d + n_scaled

# Set fixed stepsizes
mu_high = 1
mu_low = 0.1

# Create filter objects
fixed_high = af.NLMS(N,mu_high)
fixed_low = af.NLMS(N,mu_low)
adaptive = af.NLMS(N,1)

# Allocate result matrices and vectors
h_mat_high = np.zeros((N,l))
dist_vec_high = np.zeros((l,))

h_mat_low = np.zeros((N,l))
dist_vec_low = np.zeros((l,))

h_mat_adaptive = np.zeros((N,l))
dist_vec_adaptive = np.zeros((l,))

# For adaptive stepsize
beta = 0.995
x_vec = np.zeros(N,)
mu = np.zeros(l,)

# Adaptation
for i in range(l):

    # Fixed stepsize
    fixed_high.adapt(x[i,],y[i,])
    fixed_low.adapt(x[i,],y[i,])

    # Extract results
    h_mat_high[:,i] = fixed_high.h
    dist_vec_high[i] = 10*np.log10(np.linalg.norm(h-fixed_high.h)**2/h_norm)
    h_mat_low[:,i] = fixed_low.h
    dist_vec_low[i] = 10*np.log10(np.linalg.norm(h-fixed_low.h)**2/h_norm)

    # Adaptive stepsize, perfect knowledge
    # Filter memory
    x_vec[1:] = x_vec[0:-1]
    x_vec[0] = x[i]

    # Estimated output
    d_est = np.inner(x_vec,adaptive.h)

    # Smoothed errors
    if i == 0:
        e_u = (d[i] - d_est)**2
        e = (y[i] - d_est)**2
    else:
        e_u = beta*e_u + (1 - beta)*((d[i] - d_est)**2)
        e = beta*e + (1 - beta)*((y[i] - d_est)**2)

    # Set stepsize
    mu[i] = e_u/e
    adaptive.mu = mu[i]

    # Adapt with adaptive stepsize
    adaptive.adapt(x[i,],y[i,])

    # Extract results
    h_mat_adaptive[:,i] = adaptive.h
    dist_vec_adaptive[i] = 10*np.log10(np.linalg.norm(h-adaptive.h)**2/h_norm)


# Plot results

plt.subplot(212)
# Simulated Performance
plt.title('Performance')
plt.plot(np.linspace(0,l-1,l)/256,dist_vec_high, c='r')
plt.plot(np.linspace(0,l-1,l)/256,dist_vec_low, c='b')
plt.plot(np.linspace(0,l-1,l)/256,dist_vec_adaptive, c='g')
# Expected performance
plt.plot([0,12],[0,-60], linestyle='dashed', c='orange')
plt.axhline(10*np.log10(mu_high/(2-mu_high)/SNR_lin/h_norm), xmin=0, xmax=8000/256, c='black',linestyle='dashed')
plt.axhline(10*np.log10(mu_low/(2-mu_low)/SNR_lin/h_norm), xmin=0, xmax=8000/256, c='black',linestyle='dashed')
plt.xlim([0,312])

plt.legend((['mu = 1','mu = 0.1','adaptive','10dB / 2N','expected steady-state']))
plt.xlabel('iterations / N')
plt.ylabel('normalized system distance in dB')

# Stepsize over time
plt.subplot(211)
plt.plot(np.linspace(0,l-1,l)/256,mu, c='g')
plt.axhline(y=mu_high, xmin=0, xmax=8000/256, c='r')
plt.axhline(y=mu_low, xmin=0, xmax=8000/256, c='b')
plt.title('Effective Stepsize')
plt.ylabel('stepsize')
plt.xlim([0,312])

plt.show(block='True')

