import numpy as np
import soundfile as sf
import matplotlib.pyplot as plt

# Our own filter class
import adaptivefilters_solution as af

# Read in audiofiles
x, samplerate = sf.read('wn_10s.wav')
n, samplerate = sf.read('meas_noise.wav')
x_power = np.mean(x**2)
n_power = np.mean(n**2)
l = len(x)

# Read in true impulse response (length 256)
h, samplerate = sf.read('h_256.wav')
h_norm = np.linalg.norm(h)**2
N = 256
N_D = 20

# Simulate system output 
d = np.convolve(x,h)
d = d[0:l]

# Scale measurement noise
# SNR with respect to excitation
SNR = 30
SNR_lin = 10**(SNR/10)
n_scaled = n*np.sqrt(x_power/n_power/SNR_lin)

# Disturbed observed signal
y = d + n_scaled

# Create filter objects
estimate = af.NLMS(N+N_D,1)
distance = af.NLMS(N,1)
error = af.NLMS(N,1)

# Init for estimate object
estimate.h[0:N_D] = np.random.normal(0,0.1,(N_D,))

# Allocate matrices for estimated vector and system distance
h_mat_estimate = np.zeros((N,l))
dist_vec_estimate = np.zeros((l,))

h_mat_distance = np.zeros((N,l))
dist_vec_distance = np.zeros((l,))

h_mat_error = np.zeros((N,l))
dist_vec_error = np.zeros((l,))

# For adaptive stepsize
beta = 0.995
beta_x = 0.995
x_vec = np.zeros(N,)
x_vec_D = np.zeros(N+N_D,)
mu_error = np.zeros(l,)
mu_distance = np.zeros(l,)
mu_estimate = np.zeros(l,)

# Delayed output
y_delayed = np.append(np.zeros(N_D,), d[:-N_D]) + n_scaled

# Adaptation
for i in range(l):

    # Adaptive stepsize, perfect knowledge
    # Filter memory
    x_vec[1:] = x_vec[0:-1]
    x_vec[0] = x[i]

    x_vec_D[1:] = x_vec_D[0:-1]
    x_vec_D[0] = x[i]

    # Estimated output
    d_est_error = np.inner(x_vec,error.h)
    d_est_distance = np.inner(x_vec,distance.h)
    d_est_estimate = np.inner(x_vec_D,estimate.h)

    # Smoothed errors
    if i == 0:
        e_u = (d[i] - d_est_error)**2
        e_error = (y[i] - d_est_error)**2

        e_distance = (y[i] - d_est_distance)**2

        e_estimate = (y_delayed[i] - d_est_estimate)**2
    else:
        e_u = beta*e_u + (1 - beta)*((d[i] - d_est_error)**2)
        e_error = beta*e_error + (1 - beta)*((y[i] - d_est_error)**2)

        e_distance = beta*e_distance + (1 - beta)*((y[i] - d_est_distance)**2)

        e_estimate = beta*e_estimate + (1 - beta)*((y_delayed[i] - d_est_estimate)**2)

    # Smoothed input
    if i == 0:
        x_smooth = x[i]**2
    else:
        x_smooth = beta_x*x_smooth + (1 - beta_x)*(x[i]**2)

    # Error equation
    mu_error[i] = e_u/e_error
    error.mu = mu_error[i]

    # Knowledge of system distance
    mu_distance[i] = min(np.linalg.norm(distance.h-h)**2*x_smooth/e_distance,1)
    distance.mu = mu_distance[i]

    # Estimate of system distance
    mu_estimate[i] = min((N+N_D)/N_D*np.linalg.norm(estimate.h[0:N_D])**2*x_smooth/e_estimate,1)
    estimate.mu = mu_estimate[i]

    # Adapt with adaptive stepsize
    error.adapt(x[i,],y[i,])
    distance.adapt(x[i,],y[i,])
    estimate.adapt(x[i,],y_delayed[i,])


    h_mat_error[:,i] = error.h
    dist_vec_error[i] = 10*np.log10(np.linalg.norm(h-error.h)**2/h_norm)

    h_mat_distance[:,i] = distance.h
    dist_vec_distance[i] = 10*np.log10(np.linalg.norm(h-distance.h)**2/h_norm)

    h_mat_estimate[:,i] = estimate.h[N_D:]
    dist_vec_estimate[i] = 10*np.log10(np.linalg.norm(h-estimate.h[N_D:])**2/h_norm)


# Plot results

# Simulated performance
plt.subplot(212)
plt.title('Performance')
plt.xlabel('iterations / N')
plt.ylabel('system distance in dB')
plt.xlim([0,312])

plt.plot(np.linspace(0,l-1,l)/256,dist_vec_estimate, c='r')
plt.plot(np.linspace(0,l-1,l)/256,dist_vec_distance, c='b')
plt.plot(np.linspace(0,l-1,l)/256,dist_vec_error, c='g')
plt.legend((['estimated system distance','measured system distance','errors available']))

# Stepsize over time
plt.subplot(211)
plt.title('Effective Stepsize')
plt.ylabel('stepsize')
plt.xlim([0,312])

plt.plot(np.linspace(0,l-1,l)/256,mu_error, c='g')
plt.plot(np.linspace(0,l-1,l)/256,mu_distance, c='b')
plt.plot(np.linspace(0,l-1,l)/256,mu_estimate, c='r')

plt.show(block='True')